# coding: utf-8


def iintersects(geometry_a, geometry_b):

    """Test to make sure that the geometry
    does not cover each other"""

    return geometry_a.intersects(geometry_b) and not geometry_a.touches(geometry_b)



# coding: utf-8
from django.contrib.contenttypes.models import ContentType
from rest_framework.views import APIView
from rest_framework.response import Response
from rest_framework.serializers import ValidationError
from .validation import TopologyChecker
from .serializers import (
    TopologyErrorSerializer,
)


class TopologyCheckView(APIView):

    def process_arguments(self, request):
        content_type_id = request.data.get('content_type_id', None)
        feature_id = request.data.get('feature_id', None)
        rules = request.data.get('rules', None)
        persist = request.data.get('persist', False)

        if not content_type_id:
            raise ValidationError('content_type_id is mandatory')

        if not rules:
            rules = '__all__'

        try:
            content_type = ContentType.objects.get(pk=content_type_id)
        except ContentType.DoesNotExist:
            raise ValidationError('content_type with id %s does not exists' % content_type_id)

        model = content_type.model_class()

        if feature_id:
            qs = model.objects.filter(pk=feature_id)
        else:
            qs = model.objects.all()

        return {
            'rules': rules,
            'model': model,
            'content_type': content_type,
            'queryset': qs,
            'persist': persist
        }

    def process_request(self, request):
        """
        Core function. all the action happens here.
        Gets the data from the request, extracts useful
        info and feed that into the TopoChecker.
        Returns errors.
        Persist is optional.
        """
        data = self.process_arguments(request)
        queryset = data.get('queryset')
        rules = data.get('rules')
        persist = data.get('persist')
        topo_checker = TopologyChecker(persist=persist)
        errors = topo_checker.validate(queryset, rules)
        serializer = TopologyErrorSerializer(errors, many=True)
        return Response(serializer.data)

    def post(self, request, format='json'):
        """
        checks errors and
        persists them into the database
        """
        return self.process_request(request)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.2 on 2017-06-28 19:01
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='Rule',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=64, verbose_name='Name')),
                ('description', models.CharField(max_length=256, null=True, verbose_name='Description')),
                ('message', models.CharField(max_length=256, verbose_name='Message')),
                ('method', models.CharField(max_length=256, verbose_name='Method')),
            ],
            options={
                'verbose_name': 'Rule',
                'verbose_name_plural': 'Rules',
            },
        ),
        migrations.CreateModel(
            name='TopologyError',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('object_id_a', models.PositiveIntegerField()),
                ('object_id_b', models.PositiveIntegerField(null=True)),
                ('content_type_a', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='errors_a', to='contenttypes.ContentType')),
                ('content_type_b', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='errors_b', to='contenttypes.ContentType')),
            ],
            options={
                'verbose_name': 'Topology Error',
                'verbose_name_plural': 'Topology Errors',
            },
        ),
        migrations.CreateModel(
            name='TopologyRule',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('geom_field_a', models.CharField(default='geom', max_length=128, verbose_name='Geometry Field A')),
                ('geom_field_b', models.CharField(max_length=128, null=True, verbose_name='Geometry Field B')),
                ('tolerance', models.FloatField(null=True, verbose_name='Tolerance')),
                ('raises_error', models.BooleanField(default=False, help_text='If this raises an error, the user will not be able to save a new or existing object, if it fails on validation.', verbose_name='Raises Error?')),
                ('level', models.CharField(choices=[('WARNING', 'Warning'), ('ERROR', 'Error')], default='WARNING', max_length=32, verbose_name='Error Level')),
                ('custom_message', models.CharField(help_text='Supports interpolation for placeholders model_a, model_b and rule.', max_length=512, null=True, verbose_name='Custom Message')),
                ('content_type_a', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='topology_rules_a', to='contenttypes.ContentType', verbose_name='Model A')),
                ('content_type_b', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='topology_rules_b', to='contenttypes.ContentType', verbose_name='Model B')),
                ('rule', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='topology.Rule', verbose_name='Rule')),
            ],
            options={
                'verbose_name': 'Topology Rule',
                'verbose_name_plural': 'Topology Rules',
            },
        ),
        migrations.AddField(
            model_name='topologyerror',
            name='topology_rule',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='topology.TopologyRule', verbose_name='Topology Rule'),
        ),
    ]



# coding: utf-8
from django.conf import settings


def build_contains_filter(geometry, geom_field):

    return {
        '%s__contains' % geom_field: geometry
    }


def build_iintersects_filter(geometry, geom_field):
    return {
        '%s__relate' % geom_field: (geometry, 'T********')
    }


def validation_error_class(django=False):
    """gets the validationError class"""

    if 'djangorestframework' in settings.INSTALLED_APPS and not django:
        from djangorestframework.serializers import ValidationError
    else:
        from django.core.exceptions import ValidationError

    return ValidationError



# coding: utf-8
import importlib
from django.utils.translation import ugettext_lazy as _
from django.contrib.gis.db import models
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from .utils import validation_error_class
from .choices import LEVEL_CHOICES, LEVEL_CHOICES_WARNING


class Rule(models.Model):

    name = models.CharField(
        max_length=64,
        verbose_name=_('Name')
    )

    description = models.CharField(
        max_length=256,
        verbose_name=_('Description'),
        null=True
    )

    message = models.CharField(
        max_length=256,
        verbose_name=_('Message'),
    )

    method = models.CharField(
        max_length=256,
        verbose_name=_('Method'),
    )

    def get_method(self):
        try:
            parts = self.method.split('.')
            module = '.'.join(parts[:-1])
            method_name = parts[-1]
            module = importlib.import_module(module)
            return getattr(module, method_name, None)
        except Exception as ex:
            return None

    def __str__(self):

        return self.name

    class Meta:

        verbose_name = _('Rule')
        verbose_name_plural = _('Rules')


class TopologyRule(models.Model):

    """models a topology rule
    to be applied to a layer"""

    content_type_a = models.ForeignKey(
        ContentType,
        verbose_name=_('Model A'),
        related_name='topology_rules_a'
    )

    geom_field_a = models.CharField(
        max_length=128,
        verbose_name=_('Geometry Field A'),
        default='geom'
    )

    content_type_b = models.ForeignKey(
        ContentType,
        verbose_name=_('Model B'),
        related_name='topology_rules_b',
        null=True
    )

    geom_field_b = models.CharField(
        max_length=128,
        verbose_name=_('Geometry Field B'),
        null=True
    )

    rule = models.ForeignKey(
        Rule,
        verbose_name=_('Rule')
    )

    tolerance = models.FloatField(
        verbose_name=_('Tolerance'),
        null=True
    )

    raises_error = models.BooleanField(
        verbose_name=_('Raises Error?'),
        help_text=_('If this raises an error, the user will not be able to save a new or existing object, if it fails on validation.'),  # noqa,
        default=False
    )

    level = models.CharField(
        verbose_name=('Error Level'),
        max_length=32,
        choices=LEVEL_CHOICES,
        default=LEVEL_CHOICES_WARNING,
    )

    custom_message = models.CharField(
        verbose_name=_('Custom Message'),
        max_length=512,
        help_text=_('Supports interpolation for placeholders model_a, model_b and rule.'),
        null=True
    )

    def validate(self, feature, **kwargs):

        """Validates a feature"""

        validation_method = self.rule.get_method()
        errors = validation_method(self, feature, **kwargs)
        ValidationError = validation_error_class()
        if errors and self.raises_error:
            raise ValidationError(errors[0].message)
        return errors

    def __str__(self):
        if self.content_type_a and self.content_type_b:
            return _('%(model_a)s %(rule)s %(model_b)s') % {'model_a': self.content_type_a,
                                                            'rule': self.rule,
                                                            'model_b': self.content_type_b}
        else:
            return _('%(model_a)s %(rule)s') % {'model_a': self.content_type_a,
                                                'rule': self.rule}

    class Meta:

        verbose_name = _('Topology Rule')
        verbose_name_plural = _('Topology Rules')


class TopologyError(models.Model):

    content_type_a = models.ForeignKey(
        ContentType,
        on_delete=models.CASCADE,
        related_name='errors_a'
    )

    object_id_a = models.PositiveIntegerField()

    object_a = GenericForeignKey('content_type_a', 'object_id_a')

    content_type_b = models.ForeignKey(
        ContentType,
        on_delete=models.CASCADE,
        related_name='errors_b',
        null=True
    )

    object_id_b = models.PositiveIntegerField(
        null=True
    )

    object_b = GenericForeignKey('content_type_b', 'object_id_b')

    topology_rule = models.ForeignKey(
        TopologyRule,
        verbose_name=_('Topology Rule')
    )

    geom = models.GeometryField(
        null=True
    )

    def __str__(self):
        if self.object_id_a and self.object_id_b:
            return _('%(model_a)s %(rule)s %(model_b)s') % {'model_a': self.object_id_a,
                                                            'rule': self.topology_rule.rule.message,
                                                            'model_b': self.object_id_b}
        else:
            return _('%(model_a)s %(rule)s') % {'model_a': self.object_id_a,
                                                'rule': self.topology_rule.rule.message}

    class Meta:

        verbose_name = _('Topology Error')
        verbose_name_plural = _('Topology Errors')



# coding: utf-8
import django
from .utils import (
    build_iintersects_filter,
    build_contains_filter,
)
from django.contrib.gis.db.models.query import GeoQuerySet
if django.VERSION <= (1, 9):
    from django.contrib.gis.db.models import GeoManager
else:
    from django.db.models import Manager as GeoManager


class TopologyQuerySet(GeoQuerySet):

    def contains(self, geometry, geom_field='geom'):

        if not geometry:
            return self.none()

        qs_filter = build_contains_filter(geometry, geom_field)
        return self.filter(**qs_filter)

    def contains_feature(self, feature, geom_field='geom'):
        """
        returns all the rows that contains a feature
        """
        if not geom_field:
            raise ValueError('geom_field is mandatory')

        if not feature:
            return self.none()

        geometry = getattr(feature, geom_field)

        if not geometry:
            return self.none()

        qs_filter = self.contains(geometry, geom_field)
        return qs_filter

    def iintersects(self, geometry, geom_field='geom'):
        """
        returns all rows that iintersects a geometry
        """
        if not geometry:
            return self.none()

        if not geom_field:
            raise ValueError('geom_field is mandatory')
        qs_filter = build_iintersects_filter(geometry, geom_field)
        return self.filter(**qs_filter)

    def iintersects_feature(self, feature, geom_field='geom'):

        """
        returns all the rows that iintersects a feature (or model instance)
        """
        if not geom_field:
            raise ValueError('geom_field is mandatory')

        if not feature:
            return self.none()

        geometry = getattr(feature, geom_field)

        if not geometry:
            return self.none()

        qs_filter = self.iintersects(getattr(feature, geom_field))
        if feature.pk:
            qs_filter = qs_filter.exclude(pk=feature.pk)
        return qs_filter


class TopologyManager(GeoManager):

    def get_queryset(self):
        return TopologyQuerySet(self.model, using=self._db)

    def iintersects(self, geometry, geom_field='geom'):
        return self.get_queryset().iintersects(geometry, geom_field)

    def iintersects_feature(self, feature, geom_field='geom'):
        return self.get_queryset().iintersects_feature(feature, geom_field)

    def contains(self, geometry, geom_field='geom'):
        return self.get_queryset().contains(geometry, geom_field)

    def contains_feature(self, feature, geom_field='geom'):
        return self.get_queryset().contains_feature(feature, geom_field)


# -*- coding: utf-8 -*-
# Generated by Django 1.11.2 on 2017-07-03 17:20
from __future__ import unicode_literals

import django.contrib.gis.db.models.fields
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('topology', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='topologyerror',
            name='geom',
            field=django.contrib.gis.db.models.fields.GeometryField(null=True, srid=4326),
        ),
    ]



# coding: utf-8
from rest_framework import viewsets
from .models import (
    Rule,
    TopologyRule,
    TopologyError,
)
from .serializers import (
    RuleSerializer,
    TopologyRuleSerializer,
    TopologyErrorSerializer,
)


class RuleViewSet(viewsets.ModelViewSet):

    queryset = Rule.objects.all()
    serializer_class = RuleSerializer


class TopologyRuleViewSet(viewsets.ModelViewSet):

    queryset = TopologyRule.objects.all()
    serializer_class = TopologyRuleSerializer


class TopologyErrorViewSet(viewsets.ModelViewSet):

    queryset = TopologyError.objects.all()
    serializer_class = TopologyErrorSerializer



# -*- coding: utf-8
from django.apps import AppConfig


class TopologyConfig(AppConfig):
    name = 'topology'
    from topology.io import (
        to_multi,
        to_2d,
    )
    from topology.extensions import (
        iintersects,
    )
    from django.contrib.gis.geos import (
        GEOSGeometry,
    )
    GEOSGeometry.to_multi = to_multi
    GEOSGeometry.to_2d = to_2d
    GEOSGeometry.iintersects = iintersects
    


# coding: utf-8
from django.conf.urls import include, url
from rest_framework.routers import DefaultRouter
from topology.viewsets import (
    RuleViewSet,
    TopologyRuleViewSet,
    TopologyErrorViewSet,
)
from topology.views import TopologyCheckView


router = DefaultRouter(trailing_slash=False)
router.register('rules', RuleViewSet)
router.register('topology-rules', TopologyRuleViewSet)
router.register('topology-errors', TopologyErrorViewSet)

urlpatterns = [
    url(r'^api/', include(router.urls)),
    url(
        r'^api/topology/check/$',
        TopologyCheckView.as_view(),
        name='topology-check'
    )
]



# coding: utf-8
import six
from django.contrib.contenttypes.models import ContentType
from .utils import build_iintersects_filter
from .signals import (
    pre_check,
    post_check,
)
from .models import (
    TopologyRule,
    TopologyError,
)


class TopologyChecker(object):

    """
    Checks topology errors for
    objects and entire models.

    Inherits the name of the
    TopologyChecker by QGIS ;)
    """

    def __init__(self, persist=True, **kwargs):

        self.persist = persist
        self.arguments = kwargs

    def get_rules(self, model, rules='__all__'):
        """
        Gets all topology rules that
        model is the validated feature
        (content_type_a)
        """
        try:
            content_type_a = ContentType.objects.get(
                app_label=model._meta.app_label,
                model=model._meta.model_name
            )
        except ContentType.DoesNotExist:
            return

        qs = TopologyRule.objects.filter(
            content_type_a=content_type_a
        )

        if isinstance(rules, six.string_types) and rules == '__all__':
            return qs
        else:
            return qs.filter(rule__name__in=rules)

    def clean_errors(self, queryset, rules):

        """
        Deletes all errors
        that refere to a set of specific rules
        and specific features
        """

        TopologyError.objects.filter(
            topology_rule__in=[rule.pk for rule in rules],
            object_id_a__in=queryset.values_list('id', flat=True)
        ).delete()

    @property
    def extra_args(self):
        return self.arguments

    def validate(self, queryset, rules='__all__'):
        """
        Validates all features in the query
        for a set of rules
        If rules = '__all__' it will validate
        all the rules for that specific model
        """

        errors = list()
        model = queryset.model
        rules = self.get_rules(model, rules)

        self.clean_errors(queryset, rules)

        pre_check.send_robust(
            sender=self,
            queryset=queryset,
            rules=rules
        )

        for rule in rules:
            for feature in queryset:
                errors.extend(rule.validate(feature, **self.extra_args))
        if self.persist:
            for error in errors:
                error.save()

        post_check.send_robust(
            self,
            queryset=queryset,
            rules=rules,
            errors=errors
        )

        return errors

    def validate_feature(self, feature, rules='__all__'):
        model = feature._meta.model
        return self.validate(
            queryset=model.objects.filter(pk=feature.pk),
            rules=rules
        )

    def validate_extent(self, model, geometry, rules='__all__', geom_field='geom'):
        qs_filter = build_iintersects_filter(geometry, geom_field)
        qs = model.objects.filter(**qs_filter)
        return self.validate(
            queryset=qs,
            rules=rules
        )

    def validate_all(self, model, rules='__all__'):
        return self.validate(
            queryset=model.objects.all(),
            rules=rules
        )



# coding: utf-8
from django.utils.translation import ugettext_lazy as _


LEVEL_CHOICES_WARNING = 'WARNING'
LEVEL_CHOICES_ERROR = 'ERROR'

LEVEL_CHOICES = (
    (LEVEL_CHOICES_WARNING, _('Warning')),
    (LEVEL_CHOICES_ERROR, _('Error')),
)



# coding: utf-8
from django.contrib.gis.db.models import Union
from .models import (
    TopologyError,
)
from .utils import (
    build_iintersects_filter,
    build_contains_filter,
)


def must_be_contained(rule, feature, **kwargs):

    """
    this rule makes sure
    that the feature is
    contained within the boundaries
    of features from the B model

    A must be contained in B
    """

    model = rule.content_type_b.model_class()
    geom_field_a = rule.geom_field_a
    geom_field_b = rule.geom_field_b

    containers = model.objects.contains_feature(feature, geom_field_b)

    if containers.count() <= 0:
        intersecting = model.objects.iintersects_feature(feature, geom_field_b).aggregate(Union(geom_field_b))
        unioned_field = '%s__union' % geom_field_b
        unioned_geom = intersecting.get(unioned_field)
        if unioned_geom:
            leftover = getattr(feature, geom_field_a).difference(unioned_geom)
        else:
            leftover = getattr(feature, geom_field_a)
        return [
            TopologyError(
                content_type_a=rule.content_type_a,
                object_id_a=feature.pk,
                topology_rule=rule,
                geom=leftover
            )
        ]

    return []


def must_not_overlap(rule, feature, **kwargs):
    """
    this rule determines that
    this feature should not overlap
    with any feature on it's own model
    A should not overlap other A
    """
    errors = list()
    model = rule.content_type_a.model_class()
    geom_field = rule.geom_field_a
    intersecting = model.objects.iintersects_feature(feature, geom_field)
    target_geometry = getattr(feature, geom_field)
    for i in intersecting:
        errors.append(
            TopologyError(
                content_type_a=rule.content_type_a,
                object_id_a=i.pk,
                topology_rule=rule,
                geom=target_geometry.intersection(getattr(i, geom_field))
            )
        )

    return errors


def must_not_overlap_with(rule, feature, **kwargs):
    """
    this rule determines that this
    feature should not overlap any
    feature of the content_type_b
    of the rule
    A should not overlap B
    """
    errors = list()
    model = rule.content_type_b.model_class()
    geom_field_a = rule.geom_field_a
    geom_field_b = rule.geom_field_b
    target_geometry = getattr(feature, geom_field_a)

    # null and empty geometries do not produce errors ;)
    if not target_geometry or target_geometry.empty:
        return []

    intersecting = model.objects.iintersects(target_geometry, geom_field_b)

    for i in intersecting:
        errors.append(
            TopologyError(
                content_type_a=rule.content_type_a,
                object_id_a=feature.pk,
                content_type_b=rule.content_type_b,
                object_id_b=i.pk,
                topology_rule=rule,
                geom=target_geometry.intersection(getattr(i, geom_field_b))
            )
        )

    return errors



# coding: utf-8
from django.contrib.gis.geos import (
    GEOSGeometry,
    MultiPolygon,
    MultiLineString,
    MultiPoint
)
from django.contrib.gis.geos.prototypes.io import wkt_w


def to_pseudo(geometry):

    return geometry.transform(900913, clone=True)


def to_multi(geometry):

    """converts a geometry
    to a multi geometry"""

    if geometry.geom_type == 'Polygon':
        return MultiPolygon(geometry.clone(), srid=geometry.srid)

    if geometry.geom_type == 'LineString':
        return MultiLineString(geometry.clone(), srid=geometry.srid)

    if geometry.geom_type == 'Point':
        return MultiPoint(geometry.clone(), srid=geometry.srid)

    return geometry.clone()


def to_2d(geometry):

    """coerces a geometry to 2d"""

    if geometry.hasz:
        wkt = wkt_w(dim=2).write(geometry).decode()
        return GEOSGeometry(wkt, srid=geometry.srid)
    return geometry.clone()



# coding: utf-8
from rest_framework import serializers
from .models import (
    Rule,
    TopologyRule,
    TopologyError,
)


class RuleSerializer(serializers.ModelSerializer):

    class Meta:
        model = Rule
        fields = (
            'id',
            'name',
            'description',
            'message',
            'method'
        )


class TopologyRuleSerializer(serializers.ModelSerializer):

    class Meta:
        model = TopologyRule
        fields = (
            'id',
            'content_type_a',
            'geom_field_a',
            'content_type_b',
            'geom_field_b',
            'rule',
            'tolerance',
            'raises_error',
            'level',
            'custom_message',
        )


class TopologyErrorSerializer(serializers.ModelSerializer):

    message = serializers.SerializerMethodField()

    def get_message(self, obj):
        return obj.__str__()

    class Meta:
        model = TopologyError
        fields = (
            'id',
            'content_type_a',
            'object_id_a',
            'content_type_b',
            'object_id_b',
            'topology_rule',
            'message',
            'geom',
        )
